import os
from financespy.transaction import parse_transaction

_months = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]

class FilesystemBackend:
    def __init__(self, folder):
        self.folder = folder

    def insert_record(self, date, record):
        with open( self.file(date), "+a") as f:
            f.write(str(record) + "\n")
            f.close()

    def records(self, date):
        
        if not os.path.exists(self.file(date)):
            return

        with open( self.file(date) ) as f:
            for line in f:
                transaction = parse_transaction( line.strip() )
                transaction.date = date
                yield transaction

    def file(self,date):
        return self.month_folder(date) + str(date.day) + ".csv"

    def month_folder(self,date):
        return self.folder + str(date.year) + "/" + _months[date.month - 1] + "/"



import pytest

from financespy import Money

def test_simple_sum():
    m1 = Money(0.1)
    m2 = Money(0.2)

    # Even if initialized with floats, the internal representation use
    # Python3 (long) int, so the test bellow will be valid
    assert Money(0.3) == m1 + m2

def test_overflow():
    m1 = Money(1e20)
    m2 = Money(100)

    # Thanks to Python3's int implementation, we can rest assured
    # overflows will never occur. It switches from a 4-byte integer representation
    # to an arbitrary-sized integer.
    assert Money("100000000000000000100") == m1 + m2

def test_difference():
    result = Money("84.6")
    assert Money(100) - Money(15.4) == result
    assert 100 - Money(15.4) == result
    assert Money(100.0) - 15.4 == result

def test_multiplication():
    assert Money(15) * 4 == Money(60)
    assert 4 * Money(15) == Money(60)
    





class Money:
    def __init__(self, value = None, cents = None):

        if value is None and cents is None:
            raise AttributeError("Invalid arguments.")

        if cents is not None:
            self._cents = cents
            return

        if( isinstance(value, str) ):
            self._set_float( float(value) )
        elif( isinstance(value, int) ):
            self._cents = value * 100
        elif( isinstance(value, float) ):
            self._set_float( value )
        elif( isinstance(value, Money) ):
            self._cents = value._cents

    def is_zero(self):
        return self._cents == 0

    def _set_float( self, val ):
        self._cents = round( val * 100, 2)

    def __str__( self ):
        val = round(self._cents / 100.0, 2)
        return str( val )

    def __float__(self):
         return round(self._cents / 100.0, 2)

    def __repr__( self ):
        return self.__str__()

    def __add__(self, val):
        other = val
        if not isinstance(val, Money):
            other = Money(val)
        return Money( cents = self._cents + other._cents )

    def __radd__(self, val):
        other = Money(val)
        return other.__add__(self)

    def __sub__(self, val):
        other = val
        if not isinstance(val, Money):
            other = Money(val)
        return Money( cents = self._cents - other._cents )

    def __rsub__(self, val):
        other = Money(val)
        return other.__sub__(self)

    def __mul__(self, val):
        return Money( cents = self._cents * val  )

    __rmul__ = __mul__

    def __truediv__(self, val):
        return Money( cents = self._cents / val  )

    __rtruediv__ = __truediv__

    def __eq__(self, obj):
        return isinstance(obj,Money) and obj._cents == self._cents

    def abs(self):
        return Money( cents = abs(self._cents) )

ZERO = Money(0)



import os
import financespy.categories as categories
import financespy.money as money

class Transaction:
    def __init__(self, value, description = "", categories = [ categories.Uncategorized() ] ):
        self.value = money.Money(value)
        self.categories = categories
        self.description = description

    def as_expense(self):
        new_value = -1 * self.value.abs()
        return Transaction( new_value, self.description, self.categories )

    def as_income(self):
        return Transaction( self.value.abs(), self.description, self.categories )

    def __repr__(self):
        return ("%s, %s, %s" % ( str(self.value), self.description, str(self.categories)[1:-1]))

    def main_category(self):
        return self.categories[0]

    def test_categories(self, category):
        return [ category.is_of_category(c) for c in self.categories ]

    def is_of_category(self,category):
        for c in self.categories:
            if category.is_of_category(c):
                return True

        return False

    def __getattr__(self, name):
        if name.startswith("is_"):
            cat_name = name[3:]
            return self.is_of_category(categories.get_category(cat_name))

    __str__ = __repr__

def parse_transaction( record, separator = "," ):
    if( isinstance( record, str ) ):
        values = record.split(separator)
    elif( isinstance( record, list ) ):
        values = record
    else:
        raise Exception( str(type(record)) + " not allowed as parameter." )

    values = [ s.strip() for s in values if s.strip() ]

    if len(values) == 2:
        values.append(values[1])

    value = money.Money(values[0])
    description = values[1]

    category_list = list(set([ categories.get_category(s) for s in values[2:] ]))

    return Transaction(value, description, category_list)



#!/usr/bin/env python

import sys
import os
import datetime
import itertools
import financespy.clutil as clutil
from financespy.clutil import Command
import financespy.sh as sh

from financespy.account import Account
from financespy.filesystem_backend import FilesystemBackend
from financespy.transaction import parse_transaction


def current_year():
    return datetime.datetime.now().year

def backend(folder):
    return FilesystemBackend( folder )

@Command
def import_csv( args, workspace = "./ws/", year = current_year() ):
    month = args[0]
    file_name = args[1]

    account = Account( backend(workspace) )

    with open( file_name, "r" ) as file_:

        for line in file_:
            columns = [line.strip() for line in line.split(",") if line.strip() ]
            day = int(columns[0])
            transaction = parse_transaction(columns[1:])

            if not transaction.value.is_zero():            
                account.day( day, month, year ).insert_record( transaction )
                
@Command
def review_month( args, workspace = "./ws/", year = current_year() ):

    month = args[0]
    account = Account( backend(workspace) )

    for day in account.month( month, year ).days():
        os.system('clear')
        print( "Day: " + str(day.date) )

        records = list( day.records() )

        if not records:
            print("No expenses.")
            input()
            continue

        records_as_str = [ str(r) for r in records ]
        maxlen         = len(max( records_as_str, key=len)) if records else 0
        hr             = maxlen * "_"

        print(hr + "\n")
        print( "\n".join(records_as_str) )
        print(hr)

        values = [ r.value for r in records ]
        total = sum(values)
        print("\nTotal: " + str(total) )

        input()


@Command
def edit_day( args,  editor = "emacs", workspace = "./ws/", year = current_year() ):
    day = int(args[0])
    month = args[1]
    year = int(year)

    account = Account( backend(workspace) )

    day = account.day( day, month, year )

    file_name = day.backend.file( day.date )

    if not os.path.exists(file_name):
        sh.touch(file_name)

    sh.emacs(file_name, _bg = True)

def main():
    clutil.execute()

if __name__ == "__main__":
    main()



import sys

_categories = {}
_instances = {}

def get_category( str_ ):
    str_ = str_.lower()

    if str_ not in _categories or "uncategorized" == str_:
        return _uncategorized

    category = _categories[str_]

    if str_ not in _instances:
        _instances[str_] = category()

    return _instances[str_]



class CategoryMeta(type):
    def __init__(cls, name, parents, dct):
        _categories[name.lower()] = cls

        if hasattr( cls, "subtypes" ):
            for subtype in cls.subtypes:
                globals()[subtype] = type(subtype, (cls,), {'subtypes' : []})

        super(CategoryMeta, cls).__init__(name, parents, dct)

class Category(metaclass = CategoryMeta):
    def __str__(self):
        return self.__class__.__name__.lower()

    def is_of_category(self,cat):
        return isinstance(cat, self.__class__)

    __repr__ = __str__



class Expense(Category):
    pass

class Uncategorized(Expense):
    pass

class Income(Category):
    pass

_uncategorized = Uncategorized()


class Misc(Expense):
    pass

class Unexpected(Expense):
    pass

class Fixed(Expense):
    pass

class Food(Expense):
    pass

class Utilities(Expense):
    subtypes = [ "Internet", "Electricity", "Cellphone_Balance" ]

class Travel(Expense):
    subtypes = [ "Plane_Ticket", "Hotel_Reservation", "Train_Ticket" ]

class Leisure(Expense):
    pass

class Tax(Expense):
    subtypes = [ "TV" ]

class Shopping(Expense):
    subtypes = [ "Electronics", "Clothes", "Sports", "Home_Goods", "Furnishing", "Shopping_Misc", "Shoes", "Bag" ]

class Transport(Expense):
    subtypes = [ "Monthly_Ticket", "Single_Ticket", "Day_Ticket" ]

class Home(Expense):
    subtypes = [ "Rent", "Home_Repair", "Home_Service" ]

class Education(Expense):
    subtypes = [ "Course", "School_Book", "School_Suplies" ]

class German_Course(Course):
    pass

class Eating_Out(Food):
    subtypes = [ "Restaurant", "Fast_Food", "Icecream", "Drink", "Street_Food" ]

class Groceries(Food):
    subtypes = [ "Edeka", "Nahkauf", "Aldi" ]

class Body_And_Hygiene(Expense):
    subtypes = [ "Perfume", "Hair_Product", "Hairdresser", "Nails" ]



import datetime

def _index(a, x):
    i = bisect.bisect_left(a, x)
    if i != len(a) and a[i] == x:
        return True
    return False

class TimeFactory:
    def __init__(self, backend):
        self.backend = backend

    def month(self,month,year):
        return Month( month, year, self.backend )

    def day( self, day, month, year ):
        return self.month(month,year).day(day)

class Day:
    def __init__(self, date, backend):
        self.date  = date
        self.day = date.day
        self.backend = backend

    def __str__(self):
        return self.date.__str__()

    __repr__ = __str__

    def records(self):
        return self.backend.records( self.date )

    def insert_record(self, record):
        return self.backend.insert_record(self.date, record)

class Week:
    def __init__(self, month, days):
        self.month = month
        self._days  = days

    def days(self): 
        return self._days

    def records(self):
        for day in self._days:
            for record in day.records():
                record.date = day.date
                yield record

class Month:
    _months = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]
    _months_indexes = dict(zip(_months, range(1,13)))

    def _getMonth(m):
        if isinstance(m, str) and m in Month._months_indexes:
            m = m.lower()
            return Month._months_indexes[m]

        return int(m)

    class _MonthIterator:
        def __init__( self, m ):
            self._month = m
            self._first = datetime.date( day = 1, month = m.month, year = m.year )
            self._day   = self._first

        def __iter__(self):
            return self

        def __next__( self ):
            if self._day.month != self._first.month:
                raise StopIteration

            d = self._day
            self._day = self._day + datetime.timedelta(1)
            return Day(d, self._month.backend )


    def __init__(self, month, year, backend):
        self.month     = Month._getMonth(month)
        self.year      = year
        self.backend   = backend

    def weeks(self):
        current_week = []

        for day in self.days():
            week_day = day.date.weekday()

            if week_day == 6:
                if current_week:                
                    yield Week( self, current_week )
                    current_week = [day]
                else:
                    current_week = [day]
            else:
                current_week.append(day)

        if current_week:
            yield Week( self, current_week )

    def records(self):
        for day in self.days():
            for record in day.records():
                record.date = day.date
                yield record

    def days(self):
        return Month._MonthIterator( self )

    def day(self, day):
        try:
            return Day(datetime.date(day = day, month = self.month, year = self.year), self.backend)
        except:
            return None
        




import datetime
import financespy.time_factory as time_factory

_current_year = datetime.datetime.now().year

class Account:
    def __init__( self, backend ):
        self.timef = time_factory.TimeFactory( backend )
        
    def day(self, day, month, year = _current_year ):
        return self.timef.month(month,year).day(day)

    def month(self,month,year = _current_year):
        return self.timef.month(month, year)



import sys

_functions = {}

def _is_parameter(s):
    return s.strip().startswith("-")

def _list_to_dict(l):

    it = iter(l)
    d = {'args': []}
    key = next(it, None)

    while True:
        if key is None:
            break
        if _is_parameter(key):
            value = next(it, None)

            if value is None:
                d[key[1:]] = True
                break

            if _is_parameter(value):
                d[key[1:]] = True
                key = value
            else:
                d[key[1:]] = value
                key = next(it, None)
        else:
            d['args'].append(key)
            break

    for i in it:
        d['args'].append(i)

    return d

def Command( func ):
    _functions[func.__name__] = func
    return func

@Command
def _list_commands( args ):
    commands = list( _functions.keys() )
    commands.remove("_list_commands")
    list.sort(commands)
    print( " ".join(commands) )



def execute():
    it = iter(sys.argv[1:])
    com = next(it, None)

    if com is None:
        print("Command not found")
    else:
        if  com in _functions:
            args = _list_to_dict(it)
            func = _functions[com]
            func(** args)
        else:
            print("Command not found")
